@@ -1,8 +1,7 @@
 /**************************************************************************
  *   prompt.c  --  This file is part of GNU nano.                         *
  *                                                                        *
- *   Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007,  *
- *   2008, 2009, 2010, 2011, 2013, 2014 Free Software Foundation, Inc.    *
+ *   Copyright (C) 1999-2011, 2013-2017 Free Software Foundation, Inc.    *
  *   Copyright (C) 2016 Benno Schulenberg                                 *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
@@ -58,7 +57,7 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 	return KEY_WINCH;
 #endif
 
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
     /* If we got a mouse click and it was on a shortcut, read in the
      * shortcut character. */
     if (input == KEY_MOUSE) {
@@ -125,9 +124,9 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 	else if (s->scfunc == do_next_word_void)
 	    do_statusbar_next_word();
 #endif
-	else if (s->scfunc == do_home)
+	else if (s->scfunc == do_home_void)
 	    do_statusbar_home();
-	else if (s->scfunc == do_end)
+	else if (s->scfunc == do_end_void)
 	    do_statusbar_end();
 	/* When in restricted mode at the "Write File" prompt and the
 	 * filename isn't blank, disallow any input and deletion. */
@@ -165,7 +164,7 @@ int do_statusbar_input(bool *ran_func, bool *finished)
     return input;
 }
 
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
 /* Handle a mouse click on the statusbar prompt or the shortcut list. */
 int do_statusbar_mouse(void)
 {
@@ -197,7 +196,7 @@ void do_statusbar_output(int *the_input, size_t input_len,
 	bool filtering)
 {
     char *output = charalloc(input_len + 1);
-    char *char_buf = charalloc(mb_cur_max());
+    char onechar[MAXCHARLEN];
     int i, char_len;
 
     /* Copy the typed stuff so it can be treated. */
@@ -213,7 +212,7 @@ void do_statusbar_output(int *the_input, size_t input_len,
 	    output[i] = '\n';
 
 	/* Interpret the next multibyte character. */
-	char_len = parse_mbchar(output + i, char_buf, NULL);
+	char_len = parse_mbchar(output + i, onechar, NULL);
 
 	i += char_len;
 
@@ -225,12 +224,11 @@ void do_statusbar_output(int *the_input, size_t input_len,
 	answer = charealloc(answer, strlen(answer) + char_len + 1);
 	charmove(answer + statusbar_x + char_len, answer + statusbar_x,
 				strlen(answer) - statusbar_x + 1);
-	strncpy(answer + statusbar_x, char_buf, char_len);
+	strncpy(answer + statusbar_x, onechar, char_len);
 
 	statusbar_x += char_len;
     }
 
-    free(char_buf);
     free(output);
 
     update_the_statusbar();
@@ -296,7 +294,7 @@ void do_statusbar_cut_text(void)
     if (!ISSET(CUT_TO_END))
 	statusbar_x = 0;
 
-    null_at(&answer, statusbar_x);
+    answer[statusbar_x] = '\0';
 
     update_the_statusbar();
 }
@@ -385,35 +383,18 @@ void reinit_statusbar_x(void)
     statusbar_x = HIGHEST_POSITIVE;
 }
 
-/* Put the cursor in the answer at statusbar_x. */
-void reset_statusbar_cursor(void)
-{
-    size_t start_col = strlenpt(prompt) + 2;
-    size_t xpt = statusbar_xplustabs();
-
-    /* Work around a cursor-misplacement bug in VTEs. */
-    wmove(bottomwin, 0, 0);
-    wnoutrefresh(bottomwin);
-    doupdate();
-
-    wmove(bottomwin, 0, start_col + xpt -
-			get_statusbar_page_start(start_col, start_col + xpt));
-
-    wnoutrefresh(bottomwin);
-}
-
-/* Repaint the statusbar. */
+/* Redraw the promptbar and place the cursor at the right spot. */
 void update_the_statusbar(void)
 {
-    size_t base, the_page, end_page;
+    size_t base = strlenpt(prompt) + 2;
+    size_t the_page, end_page, column;
     char *expanded;
 
-    base = strlenpt(prompt) + 2;
     the_page = get_statusbar_page_start(base, base + strnlenpt(answer, statusbar_x));
     end_page = get_statusbar_page_start(base, base + strlenpt(answer) - 1);
 
+    /* Color the promptbar over its full width. */
     wattron(bottomwin, interface_color_pair[TITLE_BAR]);
-
     blank_statusbar();
 
     mvwaddstr(bottomwin, 0, 0, prompt);
@@ -428,14 +409,20 @@ void update_the_statusbar(void)
 
     wattroff(bottomwin, interface_color_pair[TITLE_BAR]);
 
-    reset_statusbar_cursor();
+    /* Work around a cursor-misplacement bug in VTEs. */
+    wmove(bottomwin, 0, 0);
+    wnoutrefresh(bottomwin);
+    doupdate();
+
+    /* Place the cursor at statusbar_x in the answer. */
+    column = base + statusbar_xplustabs();
+    wmove(bottomwin, 0, column - get_statusbar_page_start(base, column));
+    wnoutrefresh(bottomwin);
 }
 
 /* Get a string of input at the statusbar prompt. */
 functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
-#ifndef DISABLE_TABCOMP
 	bool allow_files, bool *listed,
-#endif
 #ifndef DISABLE_HISTORIES
 	filestruct **history_list,
 #endif
@@ -472,10 +459,6 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
 
     update_the_statusbar();
 
-    /* Refresh edit window and statusbar before getting input. */
-    wnoutrefresh(edit);
-    wnoutrefresh(bottomwin);
-
     while (TRUE) {
 	/* Ensure the cursor is shown when waiting for input. */
 	curs_set(1);
@@ -613,10 +596,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
  * The allow_tabs parameter indicates whether we should allow tabs to be
  * interpreted.  The allow_files parameter indicates whether we should
  * allow all files (as opposed to just directories) to be tab completed. */
-int do_prompt(bool allow_tabs,
-#ifndef DISABLE_TABCOMP
-	bool allow_files,
-#endif
+int do_prompt(bool allow_tabs, bool allow_files,
 	int menu, const char *curranswer,
 #ifndef DISABLE_HISTORIES
 	filestruct **history_list,
@@ -626,11 +606,10 @@ int do_prompt(bool allow_tabs,
     va_list ap;
     int retval;
     functionptrtype func = NULL;
-#ifndef DISABLE_TABCOMP
     bool listed = FALSE;
-#endif
-    /* Save a possible current statusbar x position. */
+    /* Save a possible current statusbar x position and prompt. */
     size_t was_statusbar_x = statusbar_x;
+    char *saved_prompt = prompt;
 
     bottombars(menu);
 
@@ -639,24 +618,21 @@ int do_prompt(bool allow_tabs,
 #ifndef NANO_TINY
   redo_theprompt:
 #endif
-    prompt = charalloc((COLS * mb_cur_max()) + 1);
+    prompt = charalloc((COLS * MAXCHARLEN) + 1);
     va_start(ap, msg);
-    vsnprintf(prompt, COLS * mb_cur_max(), msg, ap);
+    vsnprintf(prompt, COLS * MAXCHARLEN, msg, ap);
     va_end(ap);
     /* Reserve five columns for colon plus angles plus answer, ":<aa>". */
-    null_at(&prompt, actual_x(prompt, (COLS < 5) ? 0 : COLS - 5));
+    prompt[actual_x(prompt, (COLS < 5) ? 0 : COLS - 5)] = '\0';
 
-    func = acquire_an_answer(&retval, allow_tabs,
-#ifndef DISABLE_TABCOMP
-			allow_files, &listed,
-#endif
+    func = acquire_an_answer(&retval, allow_tabs, allow_files, &listed,
 #ifndef DISABLE_HISTORIES
 			history_list,
 #endif
 			refresh_func);
 
     free(prompt);
-    prompt = NULL;
+    prompt = saved_prompt;
 
 #ifndef NANO_TINY
     if (retval == KEY_WINCH)
@@ -763,7 +739,7 @@ int do_yesno_prompt(bool all, const char *msg)
 
 	if (func == do_cancel)
 	    response = -1;
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
 	else if (kbinput == KEY_MOUSE) {
 	    int mouse_x, mouse_y;
 	    /* We can click on the Yes/No/All shortcuts to select an answer. */
@@ -785,7 +761,7 @@ int do_yesno_prompt(bool all, const char *msg)
 		    response = -2;
 	    }
 	}
-#endif /* !DISABLE_MOUSE */
+#endif /* ENABLE_MOUSE */
 	else {
 	    /* Look for the kbinput in the Yes, No (and All) strings. */
 	    if (strchr(yesstr, kbinput) != NULL)
